/*
 * Copyright (c) <2017> <Tuomo Lehtonen, University of Helsinki>
 * 
 * 
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * 
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package aba2af;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Scanner;
import java.util.Set;

/**
 * Creates a framework/rule-graph which contains in 'sentences' the derivable
 * singletons and in 'rules' the derivable rules and in 'andNodeDerivations' the
 * "rules" used to derive the necessary and-nodes and in 'SCCs' the SCCs present
 * in this rule-graph. These four fields correspond to a rule-graph. Note: all
 * other fields are auxiliary and do not represent anything in a rule-graph and
 * should not be used in generating relevant graph sets.
 *
 * @author Tuomo Lehtonen
 */
public class Framework {

    // Derivable sentences
    private final Map<String, Sentence> sentences;

    private final HashSet<Sentence> assumptions;

    // Derivable rules
    private final List<Rule> rules;

    // "Rules" with some and-node as the head and its contents as the body
    private final List<AndNodeRule> andNodeDerivations;

    // The SCCs in this rule-graph
    private final Set<SCC> SCCs;

    private final List<String> queriedSentences;
    private final Set<Sentence> tautologicalQueried;
    private final Set<Sentence> derivableQueried;
    private final Set<Sentence> nonTrivialQueried;
    private final Set<Sentence> underivableQueried;

    private int sentenceIndex;

    static int supportSize;

    private int nAssumptionsWithContrary;

    Map<Integer, Sentence> indexToSentence;

    private int nSentencesInSCCs;
    private int nAndNodesInSCCs;

    private final Map<List<Sentence>, AndNode> andNodes;

    private final List<Rule> potentialRules;
    private final Map<Sentence, String> potentialContraries;
    private final Map<String, Sentence> potentialSentences;

    public Framework(String filename) throws FileNotFoundException {
        this.queriedSentences = new ArrayList<>();
        tautologicalQueried = new HashSet<>();
        derivableQueried = new HashSet<>();
        nonTrivialQueried = new HashSet<>();
        underivableQueried = new HashSet<>();
        sentences = new HashMap<>();
        assumptions = new HashSet<>();
        rules = new ArrayList<>();
        andNodeDerivations = new ArrayList<>();
        potentialRules = new ArrayList<>();
        andNodes = new HashMap<>();
        potentialContraries = new HashMap<>();
        potentialSentences = new HashMap<>();
        sentenceIndex = 0;
        supportSize = 0;

        Sentence truth = new Sentence("truth", false);
        potentialSentences.put("truth", truth);

        SCCs = new HashSet<>();

        buildFramework(filename);

        indexToSentence = new HashMap<>();
        for (Sentence s : sentences.values()) {
            for (Integer i : s.getSentenceIndex()) {
                indexToSentence.put(i, s);
            }
        }

        indexToSentence.put(potentialSentences.get("truth").getSentenceIndex().get(0), truth);
    }

    /*
     Scan the input file and build framework.
     */
    private void buildFramework(String filename) throws FileNotFoundException {
        List<String> ruleCandidates = new ArrayList<>();
        Scanner scanner = new Scanner(new FileReader(filename));
        while (scanner.hasNextLine()) {
            String line = scanner.nextLine();

            if (line.startsWith("myAsm")) {
                assignAssumption(line);
            } else if (line.startsWith("contrary")) {
                String attackedAssumptionName = line.split("\\(")[1].split(",")[0].replaceAll("\\s", "");
                String contraryName = line.split(",")[1].split("\\)")[0].trim();

                // Contrary won't get added as sentence yet, as it might not be derivable
                potentialContraries.put(sentences.get(attackedAssumptionName), contraryName);
            } else if (line.startsWith("myRule")) {
                ruleCandidates.add(line);
            } else if (line.startsWith("query")) {
                String query = line.split("\\(")[1].split("\\)")[0].replaceAll("\\s", "");
                queriedSentences.add(query);
            }
        }
        processRuleCandidates(ruleCandidates);
        determineRemainingDerivableRules();
        assignDerivableContraries();
        createAndNodeDerivations();
        assignSentenceIndices();
        markNodesInSCCs(new SCCDetection().detect(this));
        handleRelevantSentences();
    }

    private void handleRelevantSentences() {
        for (String s : queriedSentences) {
            Sentence sent = sentences.get(s);
            if (sent != null) {
                derivableQueried.add(sent);
                if (!tautologicalQueried.contains(sent)) {
                    nonTrivialQueried.add(sent);
                }
            } else {
                underivableQueried.add(new Sentence(s, false));
            }
        }
    }

    private void assignAssumption(String line) {
        String name = line.substring(6).split("\\)")[0].replaceAll("\\s", "");
        this.addSentence(new Sentence(name, true));
        supportSize++;
    }

    /*
     Determine if the rule is derivable at this point. If yes, add the rule, if not, store it for later.
     */
    private void processRuleCandidates(List<String> lines) {
        for (String s : lines) {
            String headName = s.substring(7).split(",")[0].replaceAll("\\s", "");
            Sentence head = processHeadOfRule(headName);
            String[] bodies = s.split("\\[")[1].split("\\]")[0].split(",");

            List<Sentence> sentencesInBody = new ArrayList<>();
            boolean derivableNow = processSentencesInBody(bodies, sentencesInBody);

            // Add rule in question to rules or potential rules.
            AndNode andNode;
            Rule rule;
            if (!andNodes.containsKey(sentencesInBody)) {
                andNode = new AndNode(sentencesInBody);
                andNodes.put(sentencesInBody, andNode);
            } else {
                andNode = andNodes.get(sentencesInBody);
            }
            rule = new Rule(head, andNode);

            if (derivableNow) {
                actualizeRule(rule);
            } else {
                this.potentialRules.add(rule);
            }
        }
    }

    private Sentence processHeadOfRule(String headName) {
        Sentence head;
        if (this.sentences.containsKey(headName)) {
            head = this.sentences.get(headName);
        } else if (potentialSentences.containsKey(headName)) {
            head = potentialSentences.get(headName);
        } else {
            head = new Sentence(headName, false);
            potentialSentences.put(headName, head);
        }
        return head;
    }

    /*
     Returns true if all of the sentences in this body of a rule are already in 'sentences'.
     Adds all the sentences not yet derivable to 'potentialSentences'.
     Creates the set with the 'Sentence' objects of this body of rule.
     */
    private boolean processSentencesInBody(String[] bodies, List<Sentence> sentencesInBody) {
        boolean derivableNow = true;
        if (!bodies[0].isEmpty()) {
            for (String s : bodies) {
                s = s.replaceAll("\\s", "");
                /* Fetch the corresponding sentence from 'sentences' or 'potentialSentences' or create it and add to
                 'potentialSentences' */
                Sentence sentenceEqualingS;
                if (this.sentences.containsKey(s)) {
                    sentenceEqualingS = this.sentences.get(s);
                } else if (potentialSentences.containsKey(s)) {
                    derivableNow = false;
                    sentenceEqualingS = potentialSentences.get(s);
                } else {
                    derivableNow = false;
                    sentenceEqualingS = new Sentence(s, false);
                    potentialSentences.put(s, sentenceEqualingS);
                }
                sentencesInBody.add(sentenceEqualingS);
            }
        } else {
            // This is to discriminate between "tautological" sentences and sentences that can't be derived.
            sentencesInBody.add(potentialSentences.get("truth"));
        }
        return derivableNow;
    }

    private void actualizeRule(Rule rule) {
        Sentence head = rule.getHead();
        AndNode body = rule.getBody();

        rules.add(rule);
        head.addChild(rule.getBody());
//        System.out.println(head.getName());
        if (queriedSentences.contains(head.getName())) {
            if (body.getContents().get(0).getName().equals("truth")) {
//                System.out.println(body.getContents().get(0).getName());
                tautologicalQueried.add(head);
            }
        }

        if (!sentences.keySet().contains(head.getName())) {
            addSentence(rule.getHead());
        }
    }

    /* Find more derivable rules iteratively until no new rule can be derived. This needs to be done 
     as some rules may be derivable after sentences are added to 'sentences' when a rule is actualized.
     */
    private void determineRemainingDerivableRules() {
        boolean changes = true;
        while (changes) {
            changes = false;
            Iterator<Rule> iter = potentialRules.iterator();
            while (iter.hasNext()) {
                Rule pRule = iter.next();
                boolean derivable = true;
                for (Sentence sent : pRule.getBody().getContents()) {
                    if (!sentences.containsValue(sent)) {
                        derivable = false;
                    }
                }
                if (derivable) {
                    actualizeRule(pRule);
                    iter.remove();
                    changes = true;
                }
            }
        }
    }

    /*
     Now that all derivable rules are determined, we create the 'rules' that are used to create the
     and nodes that are bodies of the rules.
     */
    private void createAndNodeDerivations() {
        Set<AndNode> alreadyDerived = new HashSet<>();
        for (Rule rule : this.rules) {
            AndNode andNode = rule.getBody();
            List<Sentence> sentencesInBody = andNode.getContents();
            if (sentencesInBody.size() > 0 && !alreadyDerived.contains(andNode)) {
                for (Sentence s : sentencesInBody) {
                    andNodeDerivations.add(new AndNodeRule(andNode, s));
                }
                alreadyDerived.add(andNode);
            }
        }
    }

    /* 
     After all derivable rules and therefore derivable sentences are determined, we can assign 
     the contraries that are derivable to their corresponding assumptions.
     */
    private void assignDerivableContraries() {
        for (Sentence assumption : potentialContraries.keySet()) {
            String contraryName = potentialContraries.get(assumption);
            if (this.sentences.containsKey(contraryName)) {
                Sentence contrary = this.sentences.get(contraryName);
                assumption.setContrary(contrary);
                contrary.addAssumptionThisCounters(assumption);
            }
        }
    }

    public HashSet<Sentence> getAssumptions() {
        return assumptions;
    }

    public List<Rule> getUnderivableRules() {
        return potentialRules;
    }

    public List<AndNodeRule> getAndNodeDerivations() {
        return andNodeDerivations;
    }

    public List<Rule> getRules() {
        return rules;
    }

    public Map<String, Sentence> getSentences() {
        return sentences;
    }

    private void addSentence(Sentence sentence) {
        sentences.put(sentence.getName(), sentence);
        if (sentence.isAssumption()) {
            assumptions.add(sentence);
        }
    }

    private void assignSentenceIndices() {
        List<Sentence> asmptsWithContrary = new ArrayList<>();
        for (Sentence s : sentences.values()) {
            if (s.getContrary() != null) {
                asmptsWithContrary.add(s);
            }
        }
        nAssumptionsWithContrary = asmptsWithContrary.size();
        for (Sentence asmpt : asmptsWithContrary) {
            asmpt.addSentenceIndex(sentenceIndex);
            asmpt.getContrary().addSentenceIndex(sentenceIndex + assumptions.size() + 1);
            sentenceIndex++;
        }
        for (Sentence asmpt : assumptions) {
            if (asmptsWithContrary.contains(asmpt)) {
                continue;
            }
            asmpt.addSentenceIndex(sentenceIndex);
            sentenceIndex++;
        }

        potentialSentences.get("truth").addSentenceIndex(sentenceIndex);
        sentenceIndex++;

        sentenceIndex += asmptsWithContrary.size();
        for (Sentence sentence : sentences.values()) {
            if (sentence.isAssumption() || !sentence.getAssumptionsThisCounters().isEmpty()) {
                continue;
            }
            sentence.addSentenceIndex(sentenceIndex);
            sentenceIndex++;
        }
    }

    public int getnAssumptionsWithContrary() {
        return nAssumptionsWithContrary;
    }

    @Override
    public String toString() {
        StringBuilder result = new StringBuilder("Assumptions and contraries: \n");

        for (Sentence s : this.getAssumptions()) {
            result = result.append(s).append(" -- ").append(s.getContrary()).append("\n");
        }
        result = result.append("\nRules:\n");
        for (Rule rule : rules) {
            result = result.append(rule).append("\n");
        }
        result = result.append("\nAnd-node Derivations:\n");
        for (AndNodeRule anrule : andNodeDerivations) {
            result = result.append(anrule).append("\n");
        }

        return result.toString();
    }

    public void markNodesInSCCs(Set<SCC> SCCs) {
        int sCount = 0;
        int aCount = 0;
        for (SCC scc : SCCs) {
            sCount += scc.getSentences().size();
            aCount += scc.getAndNodes().size();
        }
        nSentencesInSCCs = sCount;
        nAndNodesInSCCs = aCount;

        this.SCCs.addAll(SCCs);
    }

    public List<String> getQueriedSentences() {
        return queriedSentences;
    }

    public Set<Sentence> getTautologicalQueried() {
        return tautologicalQueried;
    }

    public Set<Sentence> getDerivableQueried() {
        return derivableQueried;
    }

    public Set<Sentence> getNonTrivialQueried() {
        return nonTrivialQueried;
    }

    public Set<Sentence> getUnderivableQueried() {
        return underivableQueried;
    }

    public int getnSentencesInSCCs() {
        return nSentencesInSCCs;
    }

    public int getnAndNodesInSCCs() {
        return nAndNodesInSCCs;
    }

    public Set<SCC> getSCCs() {
        return SCCs;
    }

    public int getSentenceIndex() {
        return sentenceIndex;
    }

    public int getSupportSize() {
        return supportSize;
    }

    public Sentence getIndexToSentence(int i) {
        return indexToSentence.get(i);
    }

}
