/*
 * Copyright (c) <2017> <Tuomo Lehtonen, University of Helsinki>
 * 
 * 
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * 
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package aba2af;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;
import org.apache.commons.cli.*;

/**
 *
 *
 * @author tuletule
 */
public class Main {

    public static Framework framework;
    public static RelevantArguments relgraphs;

    /**
     * @param args the command line arguments
     *
     */
    public static void main(String[] args) {
        Options options = new Options();

        Option sem = new Option("s", "semantic", true, "Semantic (adm | prf | stb)");
        Option reasoningTask = new Option("r", "reasoning-task", true, "Reasoning task (cred | skept)");
        Option intputFile = new Option("f", "file", true, "Input file defining the ABA framework");
        Option enumerate = new Option("e", "enumeration-mode", false,
                "If set, the program prints the AF instance file and full query files for all queried sentences, "
                + "even if they could be trivially decided");
        Option arguments = new Option("a", "arguments-file", false, "If set, a file "
                + "showing the contents of each AF argument is produced");

        sem.setRequired(true);
        reasoningTask.setRequired(true);
        intputFile.setRequired(true);

        options.addOption(sem);
        options.addOption(reasoningTask);
        options.addOption(intputFile);
        options.addOption(enumerate);
        options.addOption(arguments);
        options.addOption(new Option("h", "help", false, "Display help message and exit"));

        if (Arrays.asList(args).contains("-h") || Arrays.asList(args).contains("--help")) {
            String header = "\nProgram for translating ABA frameworks to Abstract Frameworks\n";
            String footer = "\nFor details visit https://www.cs.helsinki.fi/group/coreo/aba2af/";
            HelpFormatter formatter = new HelpFormatter();
            formatter.printHelp("aba2af", header, options, footer, true);
            return;
        }

        String semantic;
        String task;
        String file;
        boolean enumeration;
        boolean argFile;

        CommandLineParser parser = new DefaultParser();
        CommandLine line;

        try {
            line = parser.parse(options, args);

            semantic = line.getOptionValue("s");
            task = line.getOptionValue("r");
            file = line.getOptionValue("f");
            enumeration = line.hasOption("e");
            argFile = line.hasOption("a");

            if (!(semantic.equals("adm") || semantic.equals("prf") || semantic.equals("stb"))) {
                throw new ParseException("Incorrect argument for semantic. Try \"adm\" or \"prf\" or \"std\".");
            }
            if (!(task.equals("cred") || task.equals("skept"))) {
                throw new ParseException("Incorrect argument for reasoning task. Try \"cred\" or \"skept\".");
            }
        } catch (ParseException ex) {
            System.err.println("Problem parsing command line: " + ex.getMessage());
            return;
        }

        String[] splitFilename = file.split("/");
        String outputFilename = splitFilename[splitFilename.length - 1].split("\\.")[0];

        String filename = outputFilename;

        try {
            framework = new Framework(file);
        } catch (FileNotFoundException ex) {
            System.err.println("Can't read file: " + filename);
            System.err.println("Error message: " + ex.getMessage());
            return;
        }
        
        if (semantic.equals("adm") && task.equals("skept")) {
            for (Sentence s : framework.getTautologicalQueried()) {
                System.out.println("Sentence " + s.getName() + " is tautological "
                        + "and therefore skeptically accepted under admissible semantics");
            }
            for (Sentence s : framework.getNonTrivialQueried()) {
                System.out.println("Sentence " + s.getName() + " is NOT tautological "
                        + "and therefore NOT accepted skeptically under admissible semantics");
            }
            for (Sentence s : framework.getUnderivableQueried()) {
                System.out.println("Sentence " + s.getName() + " is NOT tautological "
                        + "and therefore NOT accepted skeptically under admissible semantics");
            }
            return;
        }
        
        Set<Sentence> queried = new HashSet();

        if (!enumeration) {
            queried.addAll(framework.getNonTrivialQueried());

            if (semantic.equals("adm") || semantic.equals("prf")) {
                Printing.printTrue(semantic, task, framework.getTautologicalQueried(), filename);
                Printing.printFalse(semantic, task, framework.getUnderivableQueried(), filename);
            } else if (semantic.equals("stb")) {
                if (task.equals("cred")) {
                    Printing.printFalse(semantic, task, framework.getUnderivableQueried(), filename);
                    queried.addAll(framework.getTautologicalQueried());
                } else if (task.equals("skept")) {
                    Printing.printTrue(semantic, task, framework.getTautologicalQueried(), filename);
                    queried.addAll(framework.getUnderivableQueried());
                }
            }
            if (queried.isEmpty()) {
                System.out.println("Queries trivially solvable, query files produced.");
                return;
            }
        } else {
            queried.addAll(framework.getDerivableQueried());
            queried.addAll(framework.getUnderivableQueried());
        }

        Set<Sentence> derivableQueried = new HashSet(queried);
        derivableQueried.removeAll(framework.getUnderivableQueried());

        relgraphs = new RelevantArguments(framework, derivableQueried);

        try {
            Printing.AFSolverFile(relgraphs.getGraphs(), relgraphs.getNonAttacks(), filename);
        } catch (IOException ex) {
            System.err.println("Can't write AF instance in file");
            System.err.println("Error message: " + ex.getMessage());
            return;
        }
        try {
            Printing.queries(semantic, task, filename, queried);
        } catch (IOException ex) {
            System.err.println("Can't write AF query in file");
            System.err.println("Error message: " + ex.getMessage());
            return;
        }
        if (argFile) {
            Printing.resultingArguments(filename);
        }
        System.out.println("Files succesfully produced.");
    }
}
