/*
 * Copyright (c) <2017> <Tuomo Lehtonen, University of Helsinki>
 * 
 * 
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * 
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package aba2af;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Creates the relevant argument sets associated with this framework. The field
 * graphs will hold them after this class is called. All attacks are also
 * assigned automatically. After that it is possible to print AF files
 * corresponding to different sentences of the framework.
 *
 * @author Tuomo Lehtonen
 */
public class RelevantArguments {

    // Visited nodes, or rather the ones whose all possible subgraphs have been considered
    private final Set<Node> visited;

    private final Framework framework;

    // The final relevant argument sets
    private final BitSet[] RAs;
    private final String[] RASstoString;

    // The main data-structure for holding RAS. Only one 'Graph', i.e. RAS gets created for each set of support
    private final Map<String, BitSet> supportToArgument;

    private final List<List<Integer>> nonAttackLists;
    
    // for each relevant sentence, a list of the arguments containing the sentence
    private final Map<Sentence, List<Integer>> argsContainingQueriedSentences;

    private final Set<Sentence> queriedSentences;

    public RelevantArguments(Framework framework, Set<Sentence> queriedSentences) {
        this.framework = framework;
        visited = new HashSet<>();
        supportToArgument = new HashMap<>();

        createRelevantArgumentSets();

        RAs = supportToArgument.values().toArray(new BitSet[supportToArgument.size()]);
        RASstoString = new String[RAs.length];
        for (int i = 0; i < RAs.length; i++) {
            RASstoString[i] = RAs[i].toString();
        }

        nonAttackLists = new ArrayList<>(RAs.length);
        for (int i = 0; i < RAs.length; i++) {
            nonAttackLists.add(new ArrayList<Integer>());
        }

        this.queriedSentences = queriedSentences;
        argsContainingQueriedSentences = new HashMap<>();
        for (Sentence queried : queriedSentences) {
            argsContainingQueriedSentences.put(queried, new ArrayList<Integer>());
        }

        queriesAndAttacks();
    }

    /*
    First iterate over all nodes in SCCs and mark them as visited. 
    Then continue to all unvisited nodes successively.
     */
    private void createRelevantArgumentSets() {
        for (SCC scc : framework.getSCCs()) {
            for (Sentence s : scc.getSentences()) {
                Set<Node> currentPath = new HashSet<>();
                orDFS(s, currentPath);
                visited.add(s);
            }
        }
        for (Sentence s : framework.getSentences().values()) {
            if (!visited.contains(s)) {
                Set<Node> currentPath = new HashSet<>();
                orDFS(s, currentPath);
            }
        }
    }

    /*
    Processes or-nodes. First go through its (grand, etc)children (that are not visited and not in current path)
    and then add this node as part of each of their partGraphs. Unless this is an assumption or truth, 
    in which case create new graph.
     */
    private void orDFS(Sentence node, Set<Node> currentPath) {
        if (!node.isInSCC()) {
            visited.add(node);
        }
        currentPath.add(node);

        if (node.isAssumption() || node.getName().equals("truth")) {
            // try a different size / no size at all
            BitSet newArgument = new BitSet(framework.getSentences().size());
            for (int index : node.getSentenceIndex()) {
                newArgument.set(index);
            }
            supportToArgument.put(newArgument.get(0, framework.getSupportSize()).toString(), newArgument);
            node.addArgumentsContainingThis(newArgument);

            currentPath.remove(node);
            return;
        }

        for (int i = 0; i < node.getChildren().size(); i++) {
            AndNode child = node.getChildren().get(i);
            if (currentPath.contains(child)) {
                continue;
            }
            if (!visited.contains(child)) {
                andDFS(child, currentPath);
            }
            for (BitSet bs : child.getArgumentsContainingThis()) {
                for (int index : node.getSentenceIndex()) {
                    bs.set(index);
                }
                node.addArgumentsContainingThis(bs);
            }
        }
        currentPath.remove(node);
    }

    /*
    Processes and-nodes. Goes through its children etc, and then calls calculateCombinations, 
    which determines all possible combinations of the this node's children's partGraphs (and of course
    adds this node to them).    
     */
    private void andDFS(AndNode node, Set<Node> currentPath) {
        if (!node.isInSCC()) {
            visited.add(node);
        } else {
            // Probably doesn't do much, but maybe saves a bit of time?
            boolean isVisited = true;
            for (Sentence s : node.getContents()) {
                if (s.isInSCC() && !visited.contains(s)) {
                    isVisited = false;
                }
            }
            if (isVisited) {
                visited.add(node);
            }
        }
        currentPath.add(node);

        List<Sentence> children = node.getContents();
        for (Sentence child : children) {
            if (currentPath.contains(child)) {
                currentPath.remove(node);
                return;
            }
            if (!visited.contains(child)) {
                orDFS(child, currentPath);
            }
        }

        calculateCombinations(children, node);

        currentPath.remove(node);
    }

    /* 
    Calculates the combinations of the node's children's partGraphs and creates those if there is not already
    a graph with the same support. If there is, the graphs get appended together.
    All elimination of redundant graphs in terms of support is done here.
     */
    private void calculateCombinations(List<Sentence> children, AndNode node) {
        Set<BitSet> combinationArgumentBitSets = new HashSet<>();

        // Process first child and add its partgraphs to the set.
        Sentence firstChild = children.get(0);
        for (BitSet bs : firstChild.getArgumentsContainingThis()) {
            BitSet newBS = (BitSet) bs.clone();
            combinationArgumentBitSets.add(newBS);
        }

        // Create all combinations by going through each previous member in the set of combinatinoGraphs.
        for (int i = 1; i < children.size(); i++) {
            Set<BitSet> newArguments = new HashSet<>();
            for (BitSet previousCombination : combinationArgumentBitSets) {
                for (BitSet partArgument : children.get(i).getArgumentsContainingThis()) {
                    BitSet newArgument = (BitSet) previousCombination.clone();
                    newArgument.or(partArgument);
                    newArguments.add(newArgument);
                }
            }
            combinationArgumentBitSets = newArguments;
        }

        // Check if there already exists a graph with the given support. If so, join them together. 
        //If not, create a new entry in the map for this set of support.
        for (BitSet bs : combinationArgumentBitSets) {
            String bsSupport = bs.get(0, framework.getSupportSize()).toString();
            if (supportToArgument.containsKey(bsSupport)) {
                BitSet existing = supportToArgument.get(bsSupport);
                existing.or(bs);
                bs = existing;
            } else {
                supportToArgument.put(bsSupport, bs);
            }
            node.addArgumentsContainingThis(bs);
        }
    }

    private void queriesAndAttacks() {
        int firstNonAsmptIndex = framework.getAssumptions().size();
        int firstAsmptWithoutContrary = framework.getnAssumptionsWithContrary();
        
        for (int i = 0; i < RAs.length; i++) {
            /* Adding the not self-attacking arguments that contain a given queried sentence to the map.
            Could be made faster by only going through the arguments that contain the given relevant argument
            IF addArgumentsContaininThis would be always updated for sentences. It isn't though.
            */
            for (Sentence sent : queriedSentences) {
                if (RAs[i].get(sent.getSentenceIndex().get(0))
                        && !RAs[i].get(firstNonAsmptIndex + 1, framework.getSentenceIndex() - 1)
                                .intersects(RAs[i].get(0, firstAsmptWithoutContrary))) {
                    argsContainingQueriedSentences.get(sent).add(i);
                }
            }

            /* Assigning attacks. Works because contraries are positioned in the same order as their 
            attacked assumptions.
            */
            for (int j = i; j < RAs.length; j++) {
                if (!RAs[i].get(firstNonAsmptIndex + 1, framework.getSentenceIndex() - 1)
                        .intersects(RAs[j].get(0, firstAsmptWithoutContrary))) {
                    nonAttackLists.get(i).add(j);
                }
                if (!RAs[j].get(firstNonAsmptIndex + 1, framework.getSentenceIndex() - 1)
                        .intersects(RAs[i].get(0, firstAsmptWithoutContrary))) {
                    nonAttackLists.get(j).add(i);
                }
            }
        }
    }

    public Map<Sentence, List<Integer>> getArgsContainingQueriedSentences() {
        return argsContainingQueriedSentences;
    }

    public List<List<Integer>> getNonAttacks() {
        return this.nonAttackLists;
    }
    
    public BitSet[] getGraphs() {
        return RAs;
    }
}
