/*
 * Copyright (c) <2017> <Tuomo Lehtonen, University of Helsinki>
 * 
 * 
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * 
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package aba2af;

import java.util.ArrayDeque;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Objects;
import java.util.Set;

/**
 * SCC detection algorithm based on Tarjan's algorithm. See for example 
 * https://en.wikipedia.org/wiki/Tarjan%27s_strongly_connected_components_algorithm
 * for details.
 * 
 * @author Tuomo Lehtonen
 */
public class SCCDetection {

    private ArrayDeque<Node> stack;
    private Map<Node, Integer> index;
    private Map<Node, Integer> lowlink;
    private int i;
    private Set<SCC> SCCs;
    private Framework framework;

    public Set<SCC> detect(Framework framework) {
        stack = new ArrayDeque<>();
        index = new HashMap<>();
        lowlink = new HashMap<>();
        i = 0;
        SCCs = new HashSet<>();
        for (Sentence s : framework.getSentences().values()) {
            if (!index.containsKey(s)) {
                SCCor(s);
            }
        }
        return SCCs;
    }

    private void SCCor(Sentence sentence) {
        index.put(sentence, i);
        lowlink.put(sentence, i);
        i++;
        stack.push(sentence);

        for (AndNode child : sentence.getChildren()) {
            if (!index.containsKey(child)) {
                SCCand(child);
                lowlink.put(sentence, Math.min(lowlink.get(sentence), lowlink.get(child)));
            } else if (stack.contains(child)) {
                lowlink.put(sentence, Math.min(lowlink.get(sentence), index.get(child)));
            }
        }

        if (Objects.equals(lowlink.get(sentence), index.get(sentence))) {
            SCC newSCC = new SCC(framework);
            while (true) {
                Node node = stack.pop();
                newSCC.addNode(node);
                if (node.equals(sentence)) {
                    break;
                }
            }
            if (!newSCC.getAndNodes().isEmpty()) {
                SCCs.add(newSCC);
                for (Sentence s : newSCC.getSentences()) {
                    s.setIsInSCC();
                }
                for (AndNode an : newSCC.getAndNodes()) {
                    an.setIsInSCC();
                }
            }
        }
    }

    private void SCCand(AndNode aNode) {
        index.put(aNode, i);
        lowlink.put(aNode, i);
        i++;
        stack.push(aNode);

        for (Sentence child : aNode.getContents()) {
            if (!index.containsKey(child)) {
                SCCor(child);
                lowlink.put(aNode, Math.min(lowlink.get(aNode), lowlink.get(child)));
            } else if (stack.contains(child)) {
                lowlink.put(aNode, Math.min(lowlink.get(aNode), index.get(child)));
            }
        }

        if (Objects.equals(lowlink.get(aNode), index.get(aNode))) {
            SCC newSCC = new SCC(framework);
            while (true) {
                Node node = stack.pop();
                newSCC.addNode(node);
                if (node.equals(aNode)) {
                    break;
                }
            }
            if (!newSCC.getSentences().isEmpty()) {
                SCCs.add(newSCC);
                for (Sentence s : newSCC.getSentences()) {
                    s.setIsInSCC();
                }
                for (AndNode an : newSCC.getAndNodes()) {
                    an.setIsInSCC();
                }
            }
        }
    }
    
}
